name = 'fissa'
path = name
version = '0.6.3'
author = "Sander Keemink & Scott Lowe"
author_email = "swkeemink@scimail.eu"
description = "A Python Library estimating somatic signals in 2-photon data"
url = "https://github.com/rochefort-lab/fissa"
